/*
  File Name     : ft602_i2c_slv.v  
  Department    : IC Design, FTDI SGP
  Author        : Do Ngoc Duong
  History       : 13 Dec. 2016 - Initial Version 

  Description   : This module performs the simple I2C Slave 
*/

module ft602_i2c_slv (
  //
  input   clk,
  input   rst_n,
  //Registers interface
  output  reg_wr,
  output  reg_rd,
  output  [7:0]   reg_adr,
  output  [127:0] reg_di,
  input           reg_rdy,
  input   [127:0] reg_do,
  //i2c interface
  input  i2c_scl,
  input  i2c_sdi,
  output i2c_sdo,
  output i2c_oe
);
//***************************
localparam DeviceAddress = 7'h0D;
// 
localparam IDLE = 4'b0001;
localparam ADDR = 4'b0010;
localparam MTWR = 4'b0100;
localparam MTRD = 4'b1000;
//****************************
reg ftdi_sync_i2c_scl;
reg ftdi_sync_i2c_sdi;
reg [3:0] int_scl;
reg [3:0] int_sdi; 
//Sync 
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
  begin 
    ftdi_sync_i2c_scl <= 'b1;
    ftdi_sync_i2c_sdi <= 'b1;
    int_scl <= 4'b1111;
    int_sdi <= 4'b1111;
  end 
  else 
  begin 
    ftdi_sync_i2c_scl <= i2c_scl;
    ftdi_sync_i2c_sdi <= i2c_sdi;
    int_scl <= {int_scl[2:0],ftdi_sync_i2c_scl};
    int_sdi <= {int_sdi[2:0],ftdi_sync_i2c_sdi};
  end 
end
//Start, Stop 
wire scl_rise;
wire scl_fall;
wire start_cond;
wire stop_cond;
wire adr_match;
wire time_cond;
wire read_cond; 
reg  start_sda;

//
always @ (posedge clk or negedge rst_n)
begin
  if (!rst_n)
    start_sda <= 1'b0;
  else if (start_cond) 
    start_sda <= 1'b0;
  else if (int_scl[0] & int_sdi[1] & (!int_sdi[0]))
    start_sda <= 1'b1;
end 
// 
assign scl_rise = int_scl[0] & (!int_scl[1]);
assign scl_fall = int_scl[1] & (!int_scl[0]);
assign start_cond = scl_fall & start_sda;   
assign stop_cond  = int_scl[0] & int_sdi[0] & (!int_sdi[1]);
//State machine  
reg  [3:0] next_state;
wire [3:0] curr_state;
//
assign curr_state = next_state;
//
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    next_state <= IDLE;
  else if (stop_cond) 
    next_state <= IDLE;
  else 
    case (curr_state)
      IDLE : next_state <= start_cond ? ADDR : IDLE;
      ADDR : next_state <= time_cond  ? (adr_match ? (read_cond ? MTRD : MTWR) 
                                                   : IDLE) : ADDR;
      MTWR : next_state <= stop_cond  ? IDLE : (start_cond ? ADDR : MTWR); 
      MTRD : next_state <= stop_cond  ? IDLE : MTRD;
      default : next_state <= IDLE; 
    endcase 
end
//Delay 
reg [3:0] scl_rise_d; 
reg [3:0] scl_fall_d; 
reg [3:0] curr_stap1;
reg [3:0] curr_stap2;
//
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
  begin 
    scl_rise_d <= 'b0;
    scl_fall_d <= 'b0;
    curr_stap1 <= 'b0;
    curr_stap2 <= 'b0;
  end 
  else 
  begin 
    scl_rise_d <= {scl_rise_d[2:0],scl_rise};
    scl_fall_d <= {scl_fall_d[2:0],scl_fall};
    curr_stap1 <= curr_state;
    curr_stap2 <= curr_stap1;
  end
end 
//Counters  
reg [3:0] sclr_cnt;
reg [3:0] sclf_cnt;
reg [4:0] byte_cnt;
//  
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    sclr_cnt <= 'b0;
  else if (start_cond | stop_cond)
    sclr_cnt <= 'b0;
  else if (scl_rise)
    if (sclr_cnt == 4'd9)
      sclr_cnt <= 'b1;
    else 
      sclr_cnt <= sclr_cnt + 1'b1;
end  
//  
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    sclf_cnt <= 'b0;
  else if (start_cond | stop_cond)
    sclf_cnt <= 'b0;
  else if (scl_fall)
    if (sclf_cnt == 4'd9)
      sclf_cnt <= 'b1;
    else 
      sclf_cnt <= sclf_cnt + 1'b1;
end  
//
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    byte_cnt <= 'b0;
  else if (start_cond | stop_cond)
    byte_cnt <= 'b0;
  else if (scl_rise_d[0] & (sclr_cnt == 4'd1)) 
    byte_cnt <= (&byte_cnt) ? byte_cnt : byte_cnt + 1'b1;
end
//
reg [7:0] shift_in;
reg [7:0] shift_out;
reg [6:0] latch_adr;
reg       latch_com;
reg       latch_ack; 
reg [7:0] latch_reg;  
//Shift Data In
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    shift_in <= 'b0;
  else if (start_cond | time_cond)
    shift_in <= 'b0;
  else if (scl_rise_d[0]&(sclr_cnt<4'd9))
    shift_in <= {shift_in[6:0],int_sdi[1]};
end
//Condition for changing state
assign time_cond = scl_fall_d[0] & (sclf_cnt == 4'd9);
assign adr_match = latch_adr == DeviceAddress;
assign read_cond = latch_com; 
//Device Address & command 
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
  begin
    latch_adr <= 'b0;
    latch_com <= 'b0;
  end 
  else if (start_cond) 
  begin
    latch_adr <= 'b0;
    latch_com <= 'b0;
  end
  else if (scl_rise_d[1]&(sclr_cnt==4'd8)&(byte_cnt==5'd1))
  begin 
    latch_adr <= shift_in[7:1];
    latch_com <= shift_in[0];
  end 
end
// Register Address 
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    latch_reg <= 'b0;
  else if (stop_cond)
    latch_reg <= 'b0;
  else if (scl_rise_d[1]&(sclr_cnt==4'd8)&(byte_cnt==5'd2)&(!latch_com))
    latch_reg <= shift_in;
end
// Latch ACK 
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    latch_ack <= 1'b0;
  else if (start_cond)
    latch_ack <= 1'b0;
  else if (scl_rise_d[1]&(sclr_cnt==4'd9))
    latch_ack <= int_sdi[2];
end  
// Write data 
wire [4:0] write_cnt;
reg [127:0] write_data;
assign write_cnt = byte_cnt - 2'd3;
//
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    write_data <= 'b0;
  else if (scl_rise_d[1]&(sclr_cnt==4'd8)&(byte_cnt>5'd2)&(byte_cnt<5'd19)&(!latch_com))
  begin
    if (write_cnt == 5'h00)
      write_data[7:0] <= shift_in;
    if (write_cnt == 5'h01)
      write_data[15:8] <= shift_in;
    if (write_cnt == 5'h02)
      write_data[23:16] <= shift_in;
    if (write_cnt == 5'h03)
      write_data[31:24] <= shift_in;
    if (write_cnt == 5'h04)
      write_data[39:32] <= shift_in;
    if (write_cnt == 5'h05)
      write_data[47:40] <= shift_in;
    if (write_cnt == 5'h06)
      write_data[55:48] <= shift_in;
    if (write_cnt == 5'h07)
      write_data[63:56] <= shift_in;
    if (write_cnt == 5'h08)
      write_data[71:64] <= shift_in;
    if (write_cnt == 5'h09)
      write_data[79:72] <= shift_in;
    if (write_cnt == 5'h0A)
      write_data[87:80] <= shift_in;
    if (write_cnt == 5'h0B)
      write_data[95:88] <= shift_in;
    if (write_cnt == 5'h0C)
      write_data[103:96] <= shift_in;
    if (write_cnt == 5'h0D)
      write_data[111:104] <= shift_in;
    if (write_cnt == 5'h0E)
      write_data[119:112] <= shift_in;
    if (write_cnt == 5'h0F)
      write_data[127:120] <= shift_in;
  end
end 
// Read Data 
wire [3:0] read_cnt;
reg [127:0] read_data;
assign read_cnt = byte_cnt - 2'd1;
//  
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    shift_out <= 'b0;
  else if (scl_fall_d[2]&(sclf_cnt==4'd9)&latch_com&(byte_cnt<5'd18))
    case (read_cnt)
      4'h0 : shift_out <= read_data[7:0];
      4'h1 : shift_out <= read_data[15:8];
      4'h2 : shift_out <= read_data[23:16];
      4'h3 : shift_out <= read_data[31:24]; 
      4'h4 : shift_out <= read_data[39:32];
      4'h5 : shift_out <= read_data[47:40];
      4'h6 : shift_out <= read_data[55:48];
      4'h7 : shift_out <= read_data[63:56];
      4'h8 : shift_out <= read_data[71:64];
      4'h9 : shift_out <= read_data[79:72];
      4'hA : shift_out <= read_data[87:80];
      4'hB : shift_out <= read_data[95:88];
      4'hC : shift_out <= read_data[103:96];
      4'hD : shift_out <= read_data[111:104];
      4'hE : shift_out <= read_data[119:112];
      4'hF : shift_out <= read_data[127:120];
    endcase 
  else if (scl_fall_d[2] &(sclf_cnt< 4'd9)&latch_com&(byte_cnt<5'd18))
    shift_out <= {shift_out[6:0],1'b0};
end
//cycle counter
reg [9:0] cycl_cnt;
wire cycl_cnten;
assign cycl_cnten = ((sclf_cnt == 4'h1) && (byte_cnt == 5'h01)); 
//
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    cycl_cnt <= 10'd0;
  else if (start_cond)
    cycl_cnt <= 10'd0;
  else if (cycl_cnten)
    cycl_cnt <= cycl_cnt + 1'b1;
end
// 
reg [9:0] hold_cnt;
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    hold_cnt <= 10'd0;
  else if (scl_fall)
    hold_cnt <= 10'd0;
  else if ((sclf_cnt == 4'd9) & (hold_cnt <cycl_cnt) & (!i2c_oe))
    hold_cnt <= hold_cnt + 1'b1;
end 
// ACK 
reg int_ack;
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    int_ack <= 1'b0;
  else if (scl_fall_d[1]&(sclf_cnt==4'd8)) 
    if (byte_cnt==5'b1)
      int_ack <= !adr_match;
    else if (byte_cnt < 5'd18)
      int_ack <= 1'b0;
    else 
      int_ack <= 1'b0;
end
//SDO 
reg int_sdo;
assign i2c_sdo = int_sdo;
//
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    int_sdo <= 1'b0;
  else if (scl_fall_d[3]) 
    int_sdo <= (sclf_cnt == 4'd8) ? int_ack : shift_out[7];
end
//Output Enable 
reg int_oe;
assign i2c_oe = int_oe;
//
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    int_oe <= 1'b0;
  else if (curr_state == IDLE)
    int_oe <= 1'b0;
  else if ((curr_state == ADDR) || (curr_state == MTWR))
    if (scl_fall_d[2]) 
      int_oe <= (sclf_cnt==4'd8);
    else 
      int_oe <= int_oe;
  else if (curr_state == MTRD)
    if (scl_fall_d[1])
      int_oe <= 1'b0;
    else if (scl_fall_d[2])
      int_oe <= (sclf_cnt < 4'd8);
    else if (sclf_cnt == 4'd9)
      if (latch_ack)
        int_oe <= 1'b0;
      else          
        int_oe <= (hold_cnt >= (cycl_cnt - 3'd5)); 
    else 
      int_oe <= int_oe;
end 
//Register Interface 
assign reg_wr  = stop_cond & (!latch_com) & adr_match & (byte_cnt > 5'd1);
assign reg_rd  = scl_fall_d[0] & (byte_cnt == 5'b1) & (sclf_cnt==4'd9) & latch_com;
assign reg_adr = latch_reg;
assign reg_di  = write_data;
// 
always @ (posedge clk or negedge rst_n)
begin 
  if (!rst_n)
    read_data <= 'b0;
  else if (reg_rdy)
    read_data <= reg_do;
end 
// 
endmodule 
